/*
 * linux/fs/ldiskfs/crypto_policy.c
 *
 * Copyright (C) 2015, Google, Inc.
 *
 * This contains encryption policy functions for ldiskfs
 *
 * Written by Michael Halcrow, 2015.
 */

#include <linux/random.h>
#include <linux/string.h>
#include <linux/types.h>

#include "ldiskfs_jbd2.h"
#include "ldiskfs.h"
#include "xattr.h"

static int ldiskfs_inode_has_encryption_context(struct inode *inode)
{
	int res = ldiskfs_xattr_get(inode, LDISKFS_XATTR_INDEX_ENCRYPTION,
				 LDISKFS_XATTR_NAME_ENCRYPTION_CONTEXT, NULL, 0);
	return (res > 0);
}

/*
 * check whether the policy is consistent with the encryption context
 * for the inode
 */
static int ldiskfs_is_encryption_context_consistent_with_policy(
	struct inode *inode, const struct ldiskfs_encryption_policy *policy)
{
	struct ldiskfs_encryption_context ctx;
	int res = ldiskfs_xattr_get(inode, LDISKFS_XATTR_INDEX_ENCRYPTION,
				 LDISKFS_XATTR_NAME_ENCRYPTION_CONTEXT, &ctx,
				 sizeof(ctx));
	if (res != sizeof(ctx))
		return 0;
	return (memcmp(ctx.master_key_descriptor, policy->master_key_descriptor,
			LDISKFS_KEY_DESCRIPTOR_SIZE) == 0 &&
		(ctx.flags ==
		 policy->flags) &&
		(ctx.contents_encryption_mode ==
		 policy->contents_encryption_mode) &&
		(ctx.filenames_encryption_mode ==
		 policy->filenames_encryption_mode));
}

static int ldiskfs_create_encryption_context_from_policy(
	struct inode *inode, const struct ldiskfs_encryption_policy *policy)
{
	struct ldiskfs_encryption_context ctx;
	handle_t *handle;
	int res, res2;

	res = ldiskfs_convert_inline_data(inode);
	if (res)
		return res;

	ctx.format = LDISKFS_ENCRYPTION_CONTEXT_FORMAT_V1;
	memcpy(ctx.master_key_descriptor, policy->master_key_descriptor,
	       LDISKFS_KEY_DESCRIPTOR_SIZE);
	if (!ldiskfs_valid_contents_enc_mode(policy->contents_encryption_mode)) {
		printk(KERN_WARNING
		       "%s: Invalid contents encryption mode %d\n", __func__,
			policy->contents_encryption_mode);
		return -EINVAL;
	}
	if (!ldiskfs_valid_filenames_enc_mode(policy->filenames_encryption_mode)) {
		printk(KERN_WARNING
		       "%s: Invalid filenames encryption mode %d\n", __func__,
			policy->filenames_encryption_mode);
		return -EINVAL;
	}
	if (policy->flags & ~LDISKFS_POLICY_FLAGS_VALID)
		return -EINVAL;
	ctx.contents_encryption_mode = policy->contents_encryption_mode;
	ctx.filenames_encryption_mode = policy->filenames_encryption_mode;
	ctx.flags = policy->flags;
	BUILD_BUG_ON(sizeof(ctx.nonce) != LDISKFS_KEY_DERIVATION_NONCE_SIZE);
	get_random_bytes(ctx.nonce, LDISKFS_KEY_DERIVATION_NONCE_SIZE);

	handle = ldiskfs_journal_start(inode, LDISKFS_HT_MISC,
				    ldiskfs_jbd2_credits_xattr(inode));
	if (IS_ERR(handle))
		return PTR_ERR(handle);
	res = ldiskfs_xattr_set(inode, LDISKFS_XATTR_INDEX_ENCRYPTION,
			     LDISKFS_XATTR_NAME_ENCRYPTION_CONTEXT, &ctx,
			     sizeof(ctx), 0);
	if (!res) {
		ldiskfs_set_inode_flag(inode, LDISKFS_INODE_ENCRYPT);
		res = ldiskfs_mark_inode_dirty(handle, inode);
		if (res)
			LDISKFS_ERROR_INODE(inode, "Failed to mark inode dirty");
	}
	res2 = ldiskfs_journal_stop(handle);
	if (!res)
		res = res2;
	return res;
}

int ldiskfs_process_policy(const struct ldiskfs_encryption_policy *policy,
			struct inode *inode)
{
	if (!inode_owner_or_capable(inode))
		return -EACCES;

	if (policy->version != 0)
		return -EINVAL;

	if (!ldiskfs_inode_has_encryption_context(inode)) {
		if (!S_ISDIR(inode->i_mode))
			return -EINVAL;
		if (!ldiskfs_empty_dir(inode))
			return -ENOTEMPTY;
		return ldiskfs_create_encryption_context_from_policy(inode,
								  policy);
	}

	if (ldiskfs_is_encryption_context_consistent_with_policy(inode, policy))
		return 0;

	printk(KERN_WARNING "%s: Policy inconsistent with encryption context\n",
	       __func__);
	return -EINVAL;
}

int ldiskfs_get_policy(struct inode *inode, struct ldiskfs_encryption_policy *policy)
{
	struct ldiskfs_encryption_context ctx;

	int res = ldiskfs_xattr_get(inode, LDISKFS_XATTR_INDEX_ENCRYPTION,
				 LDISKFS_XATTR_NAME_ENCRYPTION_CONTEXT,
				 &ctx, sizeof(ctx));
	if (res != sizeof(ctx))
		return -ENOENT;
	if (ctx.format != LDISKFS_ENCRYPTION_CONTEXT_FORMAT_V1)
		return -EINVAL;
	policy->version = 0;
	policy->contents_encryption_mode = ctx.contents_encryption_mode;
	policy->filenames_encryption_mode = ctx.filenames_encryption_mode;
	policy->flags = ctx.flags;
	memcpy(&policy->master_key_descriptor, ctx.master_key_descriptor,
	       LDISKFS_KEY_DESCRIPTOR_SIZE);
	return 0;
}

int ldiskfs_is_child_context_consistent_with_parent(struct inode *parent,
						 struct inode *child)
{
	const struct ldiskfs_crypt_info *parent_ci, *child_ci;
	struct ldiskfs_encryption_context parent_ctx, child_ctx;
	int res;

	/* No restrictions on file types which are never encrypted */
	if (!S_ISREG(child->i_mode) && !S_ISDIR(child->i_mode) &&
	    !S_ISLNK(child->i_mode))
		return 1;

	/* No restrictions if the parent directory is unencrypted */
	if (!ldiskfs_encrypted_inode(parent))
		return 1;

	/* Encrypted directories must not contain unencrypted files */
	if (!ldiskfs_encrypted_inode(child))
		return 0;

	/*
	 * Both parent and child are encrypted, so verify they use the same
	 * encryption policy.  Compare the fscrypt_info structs if the keys are
	 * available, otherwise retrieve and compare the fscrypt_contexts.
	 *
	 * Note that the fscrypt_context retrieval will be required frequently
	 * when accessing an encrypted directory tree without the key.
	 * Performance-wise this is not a big deal because we already don't
	 * really optimize for file access without the key (to the extent that
	 * such access is even possible), given that any attempted access
	 * already causes a fscrypt_context retrieval and keyring search.
	 *
	 * In any case, if an unexpected error occurs, fall back to "forbidden".
	 */

	res = ldiskfs_get_encryption_info(parent);
	if (res)
		return 0;
	res = ldiskfs_get_encryption_info(child);
	if (res)
		return 0;
	parent_ci = LDISKFS_I(parent)->i_crypt_info;
	child_ci = LDISKFS_I(child)->i_crypt_info;
	if (parent_ci && child_ci) {
		return memcmp(parent_ci->ci_master_key, child_ci->ci_master_key,
			      LDISKFS_KEY_DESCRIPTOR_SIZE) == 0 &&
			(parent_ci->ci_data_mode == child_ci->ci_data_mode) &&
			(parent_ci->ci_filename_mode ==
			 child_ci->ci_filename_mode) &&
			(parent_ci->ci_flags == child_ci->ci_flags);
	}

	res = ldiskfs_xattr_get(parent, LDISKFS_XATTR_INDEX_ENCRYPTION,
			     LDISKFS_XATTR_NAME_ENCRYPTION_CONTEXT,
			     &parent_ctx, sizeof(parent_ctx));
	if (res != sizeof(parent_ctx))
		return 0;

	res = ldiskfs_xattr_get(child, LDISKFS_XATTR_INDEX_ENCRYPTION,
			     LDISKFS_XATTR_NAME_ENCRYPTION_CONTEXT,
			     &child_ctx, sizeof(child_ctx));
	if (res != sizeof(child_ctx))
		return 0;

	return memcmp(parent_ctx.master_key_descriptor,
		      child_ctx.master_key_descriptor,
		      LDISKFS_KEY_DESCRIPTOR_SIZE) == 0 &&
		(parent_ctx.contents_encryption_mode ==
		 child_ctx.contents_encryption_mode) &&
		(parent_ctx.filenames_encryption_mode ==
		 child_ctx.filenames_encryption_mode) &&
		(parent_ctx.flags == child_ctx.flags);
}

/**
 * ldiskfs_inherit_context() - Sets a child context from its parent
 * @parent: Parent inode from which the context is inherited.
 * @child:  Child inode that inherits the context from @parent.
 *
 * Return: Zero on success, non-zero otherwise
 */
int ldiskfs_inherit_context(struct inode *parent, struct inode *child)
{
	struct ldiskfs_encryption_context ctx;
	struct ldiskfs_crypt_info *ci;
	int res;

	res = ldiskfs_get_encryption_info(parent);
	if (res < 0)
		return res;
	ci = LDISKFS_I(parent)->i_crypt_info;
	if (ci == NULL)
		return -ENOKEY;

	ctx.format = LDISKFS_ENCRYPTION_CONTEXT_FORMAT_V1;
	if (DUMMY_ENCRYPTION_ENABLED(LDISKFS_SB(parent->i_sb))) {
		ctx.contents_encryption_mode = LDISKFS_ENCRYPTION_MODE_AES_256_XTS;
		ctx.filenames_encryption_mode =
			LDISKFS_ENCRYPTION_MODE_AES_256_CTS;
		ctx.flags = 0;
		memset(ctx.master_key_descriptor, 0x42,
		       LDISKFS_KEY_DESCRIPTOR_SIZE);
		res = 0;
	} else {
		ctx.contents_encryption_mode = ci->ci_data_mode;
		ctx.filenames_encryption_mode = ci->ci_filename_mode;
		ctx.flags = ci->ci_flags;
		memcpy(ctx.master_key_descriptor, ci->ci_master_key,
		       LDISKFS_KEY_DESCRIPTOR_SIZE);
	}
	get_random_bytes(ctx.nonce, LDISKFS_KEY_DERIVATION_NONCE_SIZE);
	res = ldiskfs_xattr_set(child, LDISKFS_XATTR_INDEX_ENCRYPTION,
			     LDISKFS_XATTR_NAME_ENCRYPTION_CONTEXT, &ctx,
			     sizeof(ctx), 0);
	if (!res) {
		ldiskfs_set_inode_flag(child, LDISKFS_INODE_ENCRYPT);
		ldiskfs_clear_inode_state(child, LDISKFS_STATE_MAY_INLINE_DATA);
		res = ldiskfs_get_encryption_info(child);
	}
	return res;
}
